import logging
import re
import sys
import pandas as pd
import bioregistry
import shortuuid
from owlready2 import *
from gensim.parsing import strip_non_alphanum, strip_multiple_whitespaces


BASE_IRI = "http://ccb.hms.harvard.edu/t2t/"

STOP_WORDS = {'in', 'the', 'any', 'all', 'for', 'and', 'or', 'dx', 'on', 'fh', 'tx', 'only', 'qnorm', 'w', 'iqb', 's',
              'ds', 'rd', 'rdgwas', 'ICD', 'excluded', 'excluding', 'unspecified', 'certain', 'also', 'undefined',
              'ordinary', 'least', 'squares', 'FINNGEN', 'elsewhere', 'more', 'excluded', 'classified', 'classifeid',
              'unspcified', 'unspesified', 'specified', 'acquired', 'combined', 'unspeficied', 'elsewhere', 'not', 'by',
              'strict', 'wide', 'definition', 'definitions', 'confirmed', 'chapter', 'chapters', 'controls',
              'characterized', 'main', 'diagnosis', 'hospital', 'admissions', 'other', 'resulting', 'from'}

TEMPORAL_WORDS = {'age', 'time', 'times', 'date', 'initiation', 'cessation', 'progression', 'duration', 'early', 'late',
                  'later', 'trimester'}

QUANTITY_WORDS = {'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'frequently', 'per', 'hour', 'day', 'week', 'month',
                  'year', 'years', 'total', 'quantity', 'amount', 'level', 'levels', 'volume', 'count', 'counts', 'percentage',
                  'abundance', 'proportion', 'content', 'average', 'prevalence', 'mean', 'ratio'}


def normalize_list(token_list):
    normalized_token_list = []
    for token in token_list:
        normalized_token_list.append(normalize(token))
    return normalized_token_list


def normalize(token):
    """
    Normalizes a given string by converting to lower case, removing non-word characters, stop words, white space
    :param token: Text to be normalized
    :return: Normalized string
    """
    token = strip_non_alphanum(token).lower()
    token = token.replace("_", " ")
    token = " ".join(w for w in token.split() if w not in STOP_WORDS)
    token = strip_multiple_whitespaces(token)
    return token


def remove_quotes(string):
    string = string.replace("\"", "")
    string = string.replace("\'", "")
    return string


def remove_whitespace(string):
    return string.replace(' ', '')


def curie_from_iri(iri):
    return bioregistry.curie_from_iri(iri)


def label_from_iri(iri):
    if "#" in iri:
        return iri.split("#")[1]
    else:
        return iri.rsplit('/', 1)[1]


def iri_from_tag(source_tag):
    iri = source_tag
    if len(source_tag) > 0 and source_tag != "NA":
        iri = remove_whitespace(iri)
        if ":" in source_tag:
            onto_name = iri.split(":")[0]
            term_name = iri.replace(":", "_")
            full_iri = _get_iri(onto_name, term_name)
            iri = full_iri if len(full_iri) > 0 else iri
        elif "_" in source_tag:
            onto_name = iri.split("_")[0]
            full_iri = _get_iri(onto_name, iri)
            iri = full_iri if len(full_iri) > 0 else iri
    return iri


def _get_iri(ont_name, term_name):
    iri = ''
    if ont_name in ONTOLOGY_IRIS:
        if ont_name == 'ORPHA':
            iri = ONTOLOGY_IRIS.get(ont_name) + term_name.replace('ORPHA_', 'Orphanet_')
        elif ont_name == 'SNOMED' or ont_name == 'OMIM':
            iri = ONTOLOGY_IRIS.get(ont_name) + term_name.replace(ont_name + '_', '')
        else:
            iri = ONTOLOGY_IRIS.get(ont_name) + term_name
    return iri


def get_logger(name, level):
    formatter = logging.Formatter("%(asctime)s %(levelname)s [%(name)s]: %(message)s", "%Y-%m-%d %H:%M:%S")
    logger = logging.getLogger(name)
    logger.setLevel(level=level)
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    if not logger.hasHandlers():
        logger.addHandler(console_handler)
    logger.propagate = False
    return logger


def parse_list_file(file_path):
    file = open(file_path)
    lines = file.read().splitlines()
    file.close()
    return lines


def parse_csv_file(file_path, term_column_name, term_id_column_name, separator=','):
    data = pd.read_csv(file_path, sep=separator)
    if term_column_name not in data.columns:
        raise ValueError("Could not find specified column name for input terms: " + term_column_name)
    terms = data[term_column_name].values
    if term_id_column_name not in data.columns:
        term_ids = generate_iris(len(terms))
    elif data[term_id_column_name].isnull().values.all():
        term_ids = generate_iris(len(terms))
    else:
        term_ids = data[term_id_column_name].values
    return terms, term_ids


def parse_tsv_file(file_path, term_column_name, term_id_column_name):
    return parse_csv_file(file_path, term_column_name, term_id_column_name, separator="\t")


def get_ontology_from_labels(term_labels):
    onto_iri = BASE_IRI + "Ontology-" + generate_uuid()
    onto = owlready2.get_ontology(onto_iri)
    onto.metadata.comment.append("Created dynamically using text2term")
    onto.metadata.comment.append(datetime.datetime.now())
    for term_label in term_labels:
        with onto:
            new_term_iri = generate_iri()
            new_term = types.new_class(new_term_iri, (Thing,))
            new_term.label = term_label
    return onto


def generate_uuid():
    return str(shortuuid.ShortUUID().random(length=10))


def generate_iri():
    return BASE_IRI + "R" + generate_uuid()


def generate_iris(quantity):
    return [generate_iri() for _ in range(quantity)]


OBO_BASE_IRI = "http://purl.obolibrary.org/obo/"
BIOPORTAL_BASE_IRI = "http://purl.bioontology.org/ontology/"
ORPHANET_IRI = "http://www.orpha.net/ORDO/"
ONTOLOGY_IRIS = {"EFO": "http://www.ebi.ac.uk/efo/",
                 "Orphanet": ORPHANET_IRI,
                 "ORPHA": ORPHANET_IRI,
                 "CL": OBO_BASE_IRI,
                 "MONDO": OBO_BASE_IRI,
                 "HP": OBO_BASE_IRI,
                 "UBERON": OBO_BASE_IRI,
                 "GO": OBO_BASE_IRI,
                 "DOID": OBO_BASE_IRI,
                 "CHEBI": OBO_BASE_IRI,
                 "OMIT": OBO_BASE_IRI,
                 "NCIT": OBO_BASE_IRI,
                 "MAXO": OBO_BASE_IRI,
                 "DRON": OBO_BASE_IRI,
                 "OAE": OBO_BASE_IRI,
                 "CIDO": OBO_BASE_IRI,
                 "OMIM": BIOPORTAL_BASE_IRI + "OMIM/",
                 "PATO": OBO_BASE_IRI,
                 "SNOMED": "http://snomed.info/id/"}



"""Provides Text2Term class"""

import os
import json
import datetime
from text2term import onto_utils
from text2term.mapper import Mapper
from text2term.term_collector import OntologyTermCollector
from text2term.term_graph_generator import TermGraphGenerator
from text2term.bioportal_mapper import BioPortalAnnotatorMapper
from text2term.syntactic_mapper import SyntacticMapper
from text2term.tfidf_mapper import TFIDFMapper
from text2term.zooma_mapper import ZoomaMapper

def map_file(input_file, target_ontology, base_iris=(), csv_columns=(), excl_deprecated=False, max_mappings=3,
             mapper=Mapper.TFIDF, min_score=0.3, output_file='', save_graphs=False, save_mappings=False,
             separator=','):
    """
    Maps the terms in the given input file to the specified target ontology.

    Parameters
    ----------
    input_file : str
        Path to input file containing 'source' terms to map to ontology terms (list of terms or CSV file)
    target_ontology : str
        Path or URL of 'target' ontology to map the source terms to. When the chosen mapper is BioPortal or Zooma,
        provide a comma-separated list of ontology acronyms (eg 'EFO,HPO') or write 'all' to search all ontologies
    base_iris : tuple
        Map only to ontology terms whose IRIs start with one of the strings given in this tuple, for example:
        ('http://www.ebi.ac.uk/efo','http://purl.obolibrary.org/obo/HP')
    csv_columns : tuple
        Name of the column that contains the terms to map, optionally followed by the name of the column that
        contains identifiers for the terms (eg 'my_terms,my_term_ids')
    separator : str
        Specifies the cell separator to be used when reading a non-comma-separated tabular file
    excl_deprecated : bool
        Exclude ontology terms stated as deprecated via `owl:deprecated true`
    mapper : mapper.Mapper
        Method used to compare source terms with ontology terms. One of: levenshtein, jaro, jarowinkler, jaccard,
        fuzzy, tfidf, zooma, bioportal
    max_mappings : int
        Maximum number of top-ranked mappings returned per source term
    min_score : float
        Minimum similarity score [0,1] for the mappings (1=exact match)
    output_file : str
        Path to desired output file for the mappings
    save_graphs : bool
        Save vis.js graphs representing the neighborhood of each ontology term
    save_mappings : bool
        Save the generated mappings to a file (specified by `output_file`)

    Returns
    ----------
    df
        Data frame containing the generated ontology mappings
    """
    source_terms, source_terms_ids = _load_data(input_file, csv_columns, separator)
    return map_terms(source_terms, target_ontology, source_terms_ids=source_terms_ids, base_iris=base_iris,
                    excl_deprecated=excl_deprecated, max_mappings=max_mappings, mapper=mapper, min_score=min_score,
                    output_file=output_file, save_graphs=save_graphs, save_mappings=save_mappings)

def map_terms(source_terms, target_ontology, base_iris=(), excl_deprecated=False, max_mappings=3, min_score=0.3,
        mapper=Mapper.TFIDF, output_file='', save_graphs=False, save_mappings=False, source_terms_ids=()):
    """
    Maps the terms in the given list to the specified target ontology.

    Parameters
    ----------
    source_terms : list
        List of 'source' terms to map to ontology terms
    target_ontology : str
        Path or URL of 'target' ontology to map the source terms to. When the chosen mapper is BioPortal or Zooma,
        provide a comma-separated list of ontology acronyms (eg 'EFO,HPO') or write 'all' to search all ontologies
    base_iris : tuple
        Map only to ontology terms whose IRIs start with one of the strings given in this tuple, for example:
        ('http://www.ebi.ac.uk/efo','http://purl.obolibrary.org/obo/HP')
    source_terms_ids : tuple
        Collection of identifiers for the given source terms
    excl_deprecated : bool
        Exclude ontology terms stated as deprecated via `owl:deprecated true`
    mapper : mapper.Mapper
        Method used to compare source terms with ontology terms. One of: levenshtein, jaro, jarowinkler, jaccard,
        fuzzy, tfidf, zooma, bioportal
    max_mappings : int
        Maximum number of top-ranked mappings returned per source term
    min_score : float
        Minimum similarity score [0,1] for the mappings (1=exact match)
    output_file : str
        Path to desired output file for the mappings
    save_graphs : bool
        Save vis.js graphs representing the neighborhood of each ontology term
    save_mappings : bool
        Save the generated mappings to a file (specified by `output_file`)

    Returns
    ----------
    df
        Data frame containing the generated ontology mappings
    """
    if len(source_terms_ids) != len(source_terms):
        source_terms_ids = onto_utils.generate_iris(len(source_terms))
    if output_file == '':
        timestamp = datetime.datetime.now().strftime("%d-%m-%YT%H-%M-%S")
        output_file = "t2t-mappings-" + timestamp + ".csv"
    if mapper in {Mapper.ZOOMA, Mapper.BIOPORTAL}:
        target_terms = '' if target_ontology.lower() == 'all' else target_ontology
    else:
        target_terms = _load_ontology(target_ontology, base_iris, excl_deprecated)
    mappings_df = _do_mapping(source_terms, source_terms_ids, target_terms, mapper, max_mappings, min_score)
    if save_mappings:
        _save_mappings(mappings_df, output_file)
    if save_graphs:
        _save_graphs(target_terms, output_file)
    return mappings_df

def _load_data(input_file_path, csv_column_names, separator):
    if len(csv_column_names) >= 1:
        term_id_col_name = ""
        if len(csv_column_names) == 2:
            term_id_col_name = csv_column_names[1]
        terms, term_ids = onto_utils.parse_csv_file(input_file_path, separator=separator,
                                                    term_column_name=csv_column_names[0],
                                                    term_id_column_name=term_id_col_name)
    else:
        terms = onto_utils.parse_list_file(input_file_path)
        term_ids = onto_utils.generate_iris(len(terms))
    return terms, term_ids

def _load_ontology(ontology, iris, exclude_deprecated):
    term_collector = OntologyTermCollector(ontology)
    onto_terms = term_collector.get_ontology_terms(base_iris=iris, exclude_deprecated=exclude_deprecated)
    if len(onto_terms) == 0:
        raise RuntimeError("Could not find any terms in the given ontology.")
    return onto_terms

def _do_mapping(source_terms, source_term_ids, ontology_terms, mapper, max_mappings, min_score):
    if mapper == Mapper.TFIDF:
        term_mapper = TFIDFMapper(ontology_terms)
        return term_mapper.map(source_terms, source_term_ids, max_mappings=max_mappings, min_score=min_score)
    elif mapper == Mapper.ZOOMA:
        term_mapper = ZoomaMapper()
        return term_mapper.map(source_terms, source_term_ids, ontologies=ontology_terms, max_mappings=max_mappings)
    elif mapper == Mapper.BIOPORTAL:
        term_mapper = BioPortalAnnotatorMapper("8f0cbe43-2906-431a-9572-8600d3f4266e")
        return term_mapper.map(source_terms, source_term_ids, ontologies=ontology_terms, max_mappings=max_mappings)
    elif mapper in {Mapper.LEVENSHTEIN, Mapper.JARO, Mapper.JARO_WINKLER, Mapper.INDEL, Mapper.FUZZY, Mapper.JACCARD}:
        term_mapper = SyntacticMapper(ontology_terms)
        return term_mapper.map(source_terms, source_term_ids, mapper, max_mappings=max_mappings)
    else:
        raise ValueError("Unsupported mapper: " + mapper)

def _save_mappings(mappings, output_file):
    if os.path.dirname(output_file):  # create output directories if needed
        os.makedirs(os.path.dirname(output_file), exist_ok=True)
    mappings.to_csv(output_file, index=False)

def _save_graphs(terms, output_file):
    term_graphs = TermGraphGenerator().graphs_dicts(terms)
    with open(output_file + "-term-graphs.json", 'w') as json_file:
        json.dump(term_graphs, json_file, indent=2)



"""Provides TermMapping and TermMappingCollection classes"""

import pandas as pd


class TermMapping:
    SRC_TERM = "Source Term"
    SRC_TERM_ID = "Source Term Id"
    TGT_TERM_LBL = "Mapped Term Label"
    TGT_TERM_IRI = "Mapped Term IRI"
    MAPPING_SCORE = "Mapping Score"

    def __init__(self, source_term, source_term_id, mapped_term_label, mapped_term_iri, mapping_score):
        self._source_term = source_term
        self._source_term_id = source_term_id
        self._mapped_term_label = mapped_term_label
        self._mapped_term_iri = mapped_term_iri
        self._mapping_score = mapping_score

    @property
    def source_term(self):
        return self._source_term

    @property
    def source_term_id(self):
        return self._source_term_id

    @property
    def mapped_term_label(self):
        return self._mapped_term_label

    @property
    def mapped_term_iri(self):
        return self._mapped_term_iri

    @property
    def mapping_score(self):
        return self._mapping_score

    def to_dict(self):
        return {
            self.SRC_TERM_ID: self.source_term_id,
            self.SRC_TERM: self.source_term,
            self.TGT_TERM_LBL: self.mapped_term_label,
            self.TGT_TERM_IRI: self.mapped_term_iri,
            self.MAPPING_SCORE: self.mapping_score
        }

    def __eq__(self, other):
        if isinstance(other, TermMapping):
            return self.source_term == other.source_term and self.mapped_term_iri == other.mapped_term_iri
        return False

    def __str__(self):
        return self.source_term + " -> " + self._mapped_term_label + " (" + self.mapped_term_iri + ")"


class TermMappingCollection:

    def __init__(self, mappings):
        self._mappings = mappings

    @property
    def mappings(self):
        return self._mappings

    def mappings_df(self):
        return pd.DataFrame([m.to_dict() for m in self.mappings])



"""Provides SyntacticMapper class"""

import logging
import time
import nltk
import rapidfuzz
from tqdm import tqdm
from text2term import onto_utils
from text2term.mapper import Mapper
from text2term.term_mapping import TermMapping, TermMappingCollection


class SyntacticMapper:

    def __init__(self, target_ontology_terms):
        """
        :param target_ontology_terms: Collection of ontology terms to be mapped against
        """
        self.logger = onto_utils.get_logger(__name__, logging.INFO)
        self.target_ontology_terms = target_ontology_terms

    def map(self, source_terms, source_terms_ids, mapper=Mapper.JARO_WINKLER, max_mappings=3):
        """
        :param source_terms: List of source terms to be mapped with ontology terms
        :param source_terms_ids: List of identifiers for the given source terms
        :param mapper: Mapping method to be used for matching
        :param max_mappings: Maximum number of (top scoring) ontology term mappings that should be returned
        """
        self.logger.info("Mapping %i source terms...", len(source_terms))
        start = time.time()
        mappings = []
        for term, term_id in tqdm(zip(source_terms, source_terms_ids), total=len(source_terms)):
            matches = self._map(term, term_id, mapper, max_mappings)
            mappings.extend(matches)
        end = time.time()
        self.logger.info('done (mapping time: %.2fs seconds)', end - start)
        return TermMappingCollection(mappings).mappings_df()

    def _map(self, source_term, source_term_id, mapper, max_matches=3):
        self.logger.debug("Matching %s...", source_term)
        term_matches = []
        for term in self.target_ontology_terms:
            highest_similarity = 0.0
            for target_name in self._term_names(term):
                similarity = self.compare(source_term, target_name, mapper)
                self.logger.debug("%s -> %s (%.2f)", source_term, target_name, similarity)
                if similarity > highest_similarity:
                    highest_similarity = similarity
            term_matches.append(TermMapping(source_term, source_term_id, term.label, term.iri, highest_similarity))
        matches_sorted = sorted(term_matches, key=lambda x: x.mapping_score, reverse=True)
        del matches_sorted[max_matches:]
        return matches_sorted

    def _term_names(self, ontology_term):
        lbls_syns = []
        lbls_syns.extend(ontology_term.labels)
        lbls_syns.extend(ontology_term.synonyms)
        return lbls_syns

    def compare(self, s1, s2, mapper):
        """
        Compare the given strings s1 and s2 with respect to the specified mapping method
        :param s1: source string
        :param s2: target string
        :param mapper: Mapping method to be used
        """
        if mapper == Mapper.LEVENSHTEIN:
            return self.compare_levenshtein(s1, s2)
        elif mapper == Mapper.JARO:
            return self.compare_jaro(s1, s2)
        elif mapper == Mapper.JARO_WINKLER:
            return self.compare_jarowinkler(s1, s2)
        elif mapper == Mapper.INDEL:
            return self.compare_indel(s1, s2)
        elif mapper == Mapper.FUZZY:
            return self.compare_fuzzy_ratio(s1, s2)
        elif mapper == Mapper.JACCARD:
            return self.compare_jaccard(s1, s2)
        else:
            raise ValueError("Unsupported mapping method: " + str(mapper))

    def compare_levenshtein(self, s1, s2):
        """
        Calculates the normalized Levenshtein distance between s1 and s2.
        :return similarity between s1 and s2 as a float between 0 and 1
        """
        similarity = rapidfuzz.string_metric.normalized_levenshtein(s1, s2)/100
        return similarity

    def compare_jaro(self, s1, s2):
        """
        Calculates the Jaro similarity between s1 and s2.
        :return similarity between s1 and s2 as a float between 0 and 1
        """
        similarity = rapidfuzz.string_metric.jaro_similarity(s1, s2)/100
        return similarity

    def compare_jarowinkler(self, s1, s2):
        """
        Calculates the Jaro-Winkler similarity between s1 and s2.
        :return similarity between s1 and s2 as a float between 0 and 1
        """
        similarity = rapidfuzz.string_metric.jaro_winkler_similarity(s1, s2)/100
        return similarity

    def compare_indel(self, s1, s2):
        """
        Calculates the normalized Indel distance between s1 and s2.
        See: https://maxbachmann.github.io/RapidFuzz/Usage/fuzz.html#ratio
        :return similarity between s1 and s2 as a float between 0 and 1
        """
        similarity = rapidfuzz.fuzz.ratio(s1, s2)/100
        return similarity

    def compare_fuzzy_ratio(self, s1, s2):
        """
        Calculates a weighted ratio between s1 and s2 based on rapidfuzz's fuzzy ratio algorithms.
        See: https://maxbachmann.github.io/RapidFuzz/Usage/fuzz.html#wratio
        :return similarity between s1 and s2 as a float between 0 and 1
        """
        similarity = rapidfuzz.fuzz.WRatio(s1, s2)/100
        return similarity

    def compare_jaccard(self, s1, s2):
        """
        Calculates a Jaccard-based similarity between s1 and s2.
        :return similarity between s1 and s2 as a float between 0 and 1
        """
        similarity = 1-nltk.jaccard_distance(set(s1), set(s2))
        return similarity



"""Provides Mapper enum"""

from enum import Enum


class Mapper(str, Enum):
    """ Enumeration of "mappers" (ie string similarity metrics and Web APIs) available """
    LEVENSHTEIN = 'levenshtein'
    JARO = 'jaro'
    JARO_WINKLER = 'jarowinkler'
    JACCARD = 'jaccard'
    INDEL = 'indel'
    FUZZY = 'fuzzy'
    TFIDF = 'tfidf'
    ZOOMA = 'zooma'
    BIOPORTAL = 'bioportal'

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))



import argparse
import os
import sys
from t2t import map_file
from mapper import Mapper

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='A tool for mapping free-text descriptions of (biomedical) '
                                                 'entities to controlled terms in an ontology')
    parser.add_argument("-s", "--source", required=True, type=str,
                        help="Input file containing 'source' terms to map to ontology terms (list of terms or CSV file)")
    parser.add_argument("-t", "--target", required=True, type=str,
                        help="Path or URL of 'target' ontology to map source terms to. When the chosen mapper is "
                             "BioPortal or Zooma, provide a comma-separated list of acronyms (eg 'EFO,HPO') or write "
                             "'all' to search all ontologies")
    parser.add_argument("-o", "--output", required=False, type=str, default="",
                        help="Path to desired output file for the mappings (default=current working directory)")
    parser.add_argument("-m", "--mapper", required=False, type=str, default="tfidf",
                        help="Method used to compare source terms with ontology terms. One of: " + str(Mapper.list()) +
                             " (default=tfidf)")
    parser.add_argument("-csv", "--csv_input", required=False, type=str, default=(),
                        help="Specifies that the input is a CSV file—This should be followed by the name of the column "
                             "that contains the terms to map, optionally followed by the name of the column that "
                             "contains identifiers for the terms (eg 'my_terms,my_term_ids')")
    parser.add_argument("-sep", "--separator", required=False, type=str, default=',',
                        help="Specifies the cell separator to be used when reading a non-comma-separated tabular file")
    parser.add_argument("-top", "--top_mappings", required=False, type=int, default=3,
                        help="Maximum number of top-ranked mappings returned per source term (default=3)")
    parser.add_argument("-min", "--min_score", required=False, type=float, default=0.5,
                        help="Minimum similarity score [0,1] for the mappings (1=exact match; default=0.5)")
    parser.add_argument("-iris", "--base_iris", required=False, type=str, default=(),
                        help="Map only to ontology terms whose IRIs start with a value given in this comma-separated "
                             "list (eg 'http://www.ebi.ac.uk/efo,http://purl.obolibrary.org/obo/HP)')")
    parser.add_argument("-d", "--excl_deprecated", required=False, default=False, action="store_true",
                        help="Exclude ontology terms stated as deprecated via `owl:deprecated true` (default=False)")
    parser.add_argument("-g", "--save_term_graphs", required=False, default=False, action="store_true",
                        help="Save vis.js graphs representing the neighborhood of each ontology term (default=False)")
    arguments = parser.parse_args()
    if not os.path.exists(arguments.source):
        parser.error("The file '{}' does not exist".format(arguments.source))
        sys.exit(1)
    mapper = Mapper(arguments.mapper)
    iris = arguments.base_iris
    if len(iris) > 0:
        iris = tuple(iris.split(','))
    csv_columns = arguments.csv_input
    if len(csv_columns) > 0:
        csv_columns = tuple(csv_columns.split(','))
    map_file(arguments.source, arguments.target, output_file=arguments.output, csv_columns=csv_columns,
                         excl_deprecated=arguments.excl_deprecated, mapper=mapper, max_mappings=arguments.top_mappings,
                         min_score=arguments.min_score, base_iris=iris, save_graphs=arguments.save_term_graphs,
                         save_mappings=True, separator=arguments.separator)



"""Provides TermGraph, Node and Edge classes"""


class TermGraph:
    """
    Represents a graph of the neighborhood of an ontology term.
    The graph includes all (direct and indirect) superclasses and all direct subclasses.
    """
    def __init__(self, term_iri, nodes, edges):
        self._term_iri = term_iri
        self._nodes = nodes
        self._edges = edges

    @property
    def term_iri(self):
        return self._term_iri

    @property
    def nodes(self):
        return self._nodes

    @property
    def edges(self):
        return self._edges

    def as_dict(self):
        graph = {
            "iri": self.term_iri,
            "nodes": self._nodes_dict(),
            "edges": self._edges_dict()
        }
        return graph

    def _nodes_dict(self):
        nodes = []
        for node in self.nodes:
            node = {
                "id": node.identifier,
                "label": node.label
            }
            nodes.append(node)
        return nodes

    def _edges_dict(self):
        edges = []
        for edge in self.edges:
            edge = {
                "from": edge.from_node,
                "to": edge.to_node,
                "label": edge.label
            }
            edges.append(edge)
        return edges


class Node:
    """
    Represents a node corresponding to a term in an ontology term graph.
    """
    def __init__(self, identifier, label):
        self._identifier = identifier
        self._label = label

    @property
    def identifier(self):
        return self._identifier

    @property
    def label(self):
        return self._label

    def __hash__(self):
        return hash(self.identifier)

    def __eq__(self, other):
        if isinstance(other, Node):
            return self.identifier == other.identifier
        return NotImplemented


class Edge:
    """
    Represents a labeled edge between two nodes in an ontology term graph.
    The 'from' and 'to' nodes are represented by their (IRI) identifiers.
    """
    def __init__(self, from_node, to_node, label):
        self._from_node = from_node
        self._to_node = to_node
        self._label = label

    @property
    def from_node(self):
        return self._from_node

    @property
    def to_node(self):
        return self._to_node

    @property
    def label(self):
        return self._label

    def __key(self):
        return self.from_node, self.to_node, self.label

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, Edge):
            return self.__key() == other.__key()
        return NotImplemented

    INSTANCE_OF = "INSTANCE_OF"
    IS_A = "IS_A"



"""Provides BioPortalAnnotatorMapper class"""

import json
import logging
import time
import requests
from text2term.term_mapping import TermMapping, TermMappingCollection
from text2term import onto_utils


class BioPortalAnnotatorMapper:

    def __init__(self, bp_api_key):
        """
        :param bp_api_key: BioPortal API key
        """
        self.logger = onto_utils.get_logger(__name__, logging.INFO)
        self.url = "http://data.bioontology.org/annotator"
        self.bp_api_key = bp_api_key

    def map(self, source_terms, source_terms_ids, ontologies, max_mappings=3, api_params=()):
        """
        Find and return ontology mappings through the BioPortal Annotator Web service
        :param source_terms: Collection of source terms to map to target ontologies
        :param source_terms_ids: List of identifiers for the given source terms
        :param ontologies: Comma-separated list of ontology acronyms (eg 'HP,EFO') or 'all' to search all ontologies.
            The ontology names accepted must match the names used in BioPortal. Here are some known ontologies:
            GO, UBERON, "CL" for Cell Ontology, MESH, SNOMEDCT, FMA, NCIT, EFO, DOID, MONDO, "PR" for Protein Ontology,
            "HP" for Human Phenotype Ontology
        :param max_mappings: The maximum number of (top scoring) ontology term mappings that should be returned
        :param api_params: Additional BioPortal Annotator-specific parameters to include in the request
        """
        self.logger.info("Mapping %i source terms against ontologies: %s...", len(source_terms), ontologies)
        start = time.time()
        mappings = []
        for term, term_id in zip(source_terms, source_terms_ids):
            mappings.extend(self._map_term(term, term_id, ontologies, max_mappings, api_params))
        self.logger.info('done (mapping time: %.2fs seconds)', time.time()-start)
        return TermMappingCollection(mappings).mappings_df()

    def _map_term(self, source_term, source_term_id, ontologies, max_mappings, api_params):
        params = {
            "text": onto_utils.normalize(source_term),
            "longest_only": "true",
            "expand_mappings": "true",
            "ontologies": ontologies
        }
        if len(api_params) > 0:
            params.update(api_params)
        self.logger.debug("API parameters: " + str(params))
        mappings = []
        self.logger.debug("Searching for ontology terms to match: " + source_term)
        response = self._do_get_request(self.url, params=params)
        if response is not None:
            self.logger.debug("...found " + str(len(response)) + " mappings")
            for mapping in response:
                if len(mappings) < max_mappings:
                    mappings.append(self._mapping_details(source_term, source_term_id, mapping))
        return mappings

    def _mapping_details(self, source_term, source_term_id, mapping):
        ann_class = mapping["annotatedClass"]
        term_iri = ann_class["@id"]
        term_link_bp = ann_class["links"]["self"]
        term_label = self.get_term_details(term_link_bp)
        return TermMapping(source_term, source_term_id, term_label, term_iri, 1)

    def get_term_details(self, term_iri):
        response = self._do_get_request(term_iri)
        term_label = ""
        if response is not None:
            term_label = onto_utils.remove_quotes(response["prefLabel"])
        return term_label

    def _do_get_request(self, request_url, params=None):
        headers = {
            "Authorization": "apiKey token=" + self.bp_api_key,
        }
        response = requests.get(request_url, params=params, headers=headers, verify=True)
        if response.ok:
            json_resp = json.loads(response.content)
            if len(json_resp) > 0:
                return json_resp
            else:
                self.logger.info("Empty response for input: " + request_url + " with parameters " + str(params))
        elif response.status_code == 429:  # API is throttling requests
            self.logger.info(response.reason + ". Status code: " + str(response.status_code) + ". Waiting 15 seconds.")
            time.sleep(15)
            return self._do_get_request(request_url, params)
        else:
            json_resp = json.loads(response.content)
            self.logger.error(response.reason + ":" + request_url + ". " + json_resp["errors"][0])



"""Provides OntologyTermCollector class"""

import logging
from owlready2 import *
from text2term import onto_utils
from text2term.term import OntologyTerm


class OntologyTermCollector:

    def __init__(self, ontology_iri):
        """"
        :param ontology_iri: IRI of the ontology (e.g., path of ontology document in the local file system, URL)
        """
        self.logger = onto_utils.get_logger(__name__, logging.INFO)
        self.ontology_iri = ontology_iri

    def get_ontology_terms(self, base_iris=(), use_reasoning=False, exclude_deprecated=False):
        """
        Collect the terms described in the ontology at the specified IRI
        :param base_iris: Limit ontology term collection to terms whose IRIs start with any IRI given in this tuple
        :param use_reasoning: Use a reasoner to compute inferred class hierarchy
        :param exclude_deprecated: Exclude ontology terms stated as deprecated using owl:deprecated 'true'
        :return: Collection of ontology terms in the specified ontology
        """
        ontology = self._load_ontology(self.ontology_iri)
        if use_reasoning:
            self._classify_ontology(ontology)
        self.logger.info("Collecting ontology term details...")
        start = time.time()
        ontology_terms = []
        if len(base_iris) > 0:
            for iri in base_iris:
                iri = iri.strip()
                query = iri + "*"
                self.logger.info("...collecting terms with IRIs starting in: " + iri)
                iris = list(default_world.search(iri=query))
                ontology_terms.extend(self._get_ontology_terms(iris, ontology, exclude_deprecated))
        else:
            ontology_terms = self._get_ontology_terms(ontology.classes(), ontology, exclude_deprecated)
        end = time.time()
        self.logger.info("...done: collected %i ontology terms (collection time: %.2fs)", len(ontology_terms), end-start)
        return ontology_terms

    def _get_ontology_terms(self, term_list, ontology, exclude_deprecated):
        ontology_terms = []
        for ontology_term in term_list:
            if not isinstance(ontology_term, PropertyClass) and ontology_term is not Thing and ontology_term is not Nothing:
                if (exclude_deprecated and not deprecated[ontology_term]) or (not exclude_deprecated):
                    labels = self._get_labels(ontology_term)
                    synonyms = self._get_synonyms(ontology_term)
                    parents = self._get_parents(ontology_term)
                    children = self._get_children(ontology_term, ontology)
                    instances = self._get_instances(ontology_term, ontology)
                    definition = self._get_definition(ontology_term)
                    term_details = OntologyTerm(ontology_term.iri, labels, synonyms, definition,
                                                parents=parents, children=children, instances=instances)
                    ontology_terms.append(term_details)
                else:
                    self.logger.debug("Excluding deprecated ontology term: %s", ontology_term.iri)
        return ontology_terms

    def _get_parents(self, ontology_term):
        parents = set()  # named/atomic superclasses except owl:Thing
        try:
            all_parents = ontology_term.is_a  # obtain all (direct and indirect) parents of this entity
            for parent in all_parents:
                # exclude OWL restrictions and owl:Thing and Self
                if isinstance(parent, ThingClass) and parent is not Thing and parent is not ontology_term:
                    parents.add(parent)
        except AttributeError as err:
            self.logger.debug(err)
        return parents

    def _get_children(self, ontology_term, ontology):
        children = set()
        try:
            children = set(ontology.get_children_of(ontology_term))
        except (TypeError, AttributeError) as err:
            self.logger.debug(err)
        return children

    def _get_instances(self, ontology_term, ontology):
        instances = set()
        try:
            instances = set(ontology.get_instances_of(ontology_term))
        except AttributeError as err:
            self.logger.debug(err)
        return instances

    def _get_labels(self, ontology_term):
        """
        Collect the labels of the given ontology term both given by rdfs:label and skos:prefLabel
        :param ontology_term: Ontology term
        :return: Collection of labels of the ontology term
        """
        labels = set()
        for rdfs_label in self._get_rdfs_labels(ontology_term):
            labels.add(rdfs_label)
        for skos_label in self._get_skos_pref_labels(ontology_term):
            labels.add(skos_label)
        if len(labels) == 0:
            label_from_iri = onto_utils.label_from_iri(ontology_term.iri)
            self.logger.info("...ontology term %s has no labels (rdfs:label or skos:prefLabel). "
                             "Using a label based on the term IRI: %s", ontology_term.iri, label_from_iri)
            labels.add(label_from_iri)
        self.logger.debug("...collected %i labels and synonyms for %s", len(labels), ontology_term)
        return labels

    def _get_synonyms(self, ontology_term):
        """
        Collect the synonyms of the given ontology term
        :param ontology_term: Ontology term
        :return: Collection of synonyms of the ontology term
        """
        synonyms = set()
        for synonym in self._get_obo_exact_synonyms(ontology_term):
            synonyms.add(synonym)
        for nci_synonym in self._get_nci_synonyms(ontology_term):
            synonyms.add(nci_synonym)
        for efo_alt_term in self._get_efo_alt_terms(ontology_term):
            synonyms.add(efo_alt_term)
        self.logger.debug("...collected %i synonyms for %s", len(synonyms), ontology_term)
        return synonyms

    def _get_rdfs_labels(self, ontology_term):
        """
        Collect labels of the given term that are specified using the standard rdfs:label annotation property
        :param ontology_term: Ontology term to collect labels from
        :return: Collection of RDFS labels
        """
        rdfs_labels = []
        try:
            for rdfs_label in ontology_term.label:
                rdfs_labels.append(rdfs_label)
        except AttributeError as err:
            self.logger.debug(err)
        return rdfs_labels

    def _get_skos_pref_labels(self, ontology_term):
        """
        Collect labels of the given term that are specified using the skos:prefLabel annotation property
        :param ontology_term: Ontology term to collect labels from
        :return: Collection of SKOS preferred labels
        """
        skos_labels = []
        try:
            for skos_pref_label in ontology_term.prefLabel:
                skos_labels.append(skos_pref_label)
        except AttributeError as err:
            self.logger.debug(err)
        return skos_labels

    def _get_efo_alt_terms(self, ontology_term):
        efo_alt_terms = []
        try:
            for efo_alt_term in ontology_term.alternative_term:
                efo_alt_terms.append(efo_alt_term)
        except AttributeError as err:
            self.logger.debug(err)
        return efo_alt_terms

    def _get_obo_exact_synonyms(self, ontology_term):
        """
        Collect synonyms of the given term that are specified using the annotation property used by DOID, MONDO, EFO,
        HPO, and other OBO ontologies: <http://www.geneontology.org/formats/oboInOwl#hasExactSynonym>.
        :param ontology_term: Ontology term to collect synonyms from
        :return: Collection of synonyms
        """
        synonyms = []
        try:
            for synonym in ontology_term.hasExactSynonym:
                synonyms.append(synonym)
        except AttributeError as err:
            self.logger.debug(err)
        return synonyms

    def _get_nci_synonyms(self, ontology_term):
        """
        Collect synonyms of the given term that are specified using the NCI Thesaurus annotation property:
         <http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P90>.
        :param ontology_term: Ontology term to collect synonyms from
        :return: Collection of synonyms
        """
        nci_synonyms = []
        try:
            for synonym in ontology_term.P90:
                nci_synonyms.append(synonym)
        except AttributeError as err:
            self.logger.debug(err)
        return nci_synonyms

    def _get_definition(self, ontology_term):
        """
         Get the definition (if one exists) of the given term as specified using the skos:definition annotation property
         :param ontology_term: Ontology term to collect definition of
         :return: String value of the skos:definition annotation property assertion on the given term
         """
        definition = ""
        try:
            definition = ontology_term.definition
        except AttributeError as err:
            self.logger.debug(err)
        return definition

    def _load_ontology(self, ontology_iri):
        """
        Load the ontology at the specified IRI.
        :param ontology_iri: IRI of the ontology (e.g., path of ontology document in the local file system, URL)
        :return: Ontology document
        """
        self.logger.info("Loading ontology %s...", ontology_iri)
        start = time.time()
        ontology = get_ontology(ontology_iri).load()
        end = time.time()
        self._log_ontology_metrics(ontology)
        self.logger.info("...done (ontology loading time: %.2fs)", end-start)
        return ontology

    def _classify_ontology(self, ontology):
        """
        Perform reasoning over the given ontology (consistency checking and classification)
        :param ontology: ontology instance
        """
        self.logger.info("Reasoning over ontology...")
        start = time.time()
        with ontology:  # entailments will be added to this ontology
            sync_reasoner(infer_property_values=True)
        end = time.time()
        self.logger.info("...done (reasoning time: %.2fs)", end - start)

    def _log_ontology_metrics(self, ontology):
        self.logger.debug(" Ontology IRI: %s", ontology.base_iri)
        self.logger.debug(" Class count: %i", len(list(ontology.classes())))
        self.logger.debug(" Object property count: %i", len(list(ontology.object_properties())))
        self.logger.debug(" Data property count: %i", len(list(ontology.data_properties())))
        self.logger.debug(" Annotation property count: %i", len(list(ontology.annotation_properties())))



"""Provides TFIDFMapper class"""

import logging
import time
import sparse_dot_topn as ct
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from text2term import onto_utils
from text2term.term_mapping import TermMapping, TermMappingCollection


class TFIDFMapper:

    def __init__(self, target_ontology_terms):
        """
        :param target_ontology_terms: Collection of ontology terms to be mapped against
        """
        self.logger = onto_utils.get_logger(__name__, logging.INFO)
        self.target_ontology_terms = target_ontology_terms
        self.target_labels, self.target_terms = self._get_target_labels_terms(target_ontology_terms)

    def map(self, source_terms, source_terms_ids, max_mappings=3, min_score=0.3):
        """
        Main mapping function. Default settings return only the top candidate for every source string.
        :param source_terms: List of source terms to be mapped with ontology terms
        :param source_terms_ids: List of identifiers for the given source terms
        :param max_mappings: The maximum number of (top scoring) ontology term mappings that should be returned
        :param min_score: The lower-bound threshold for keeping a candidate term mapping, between 0-1.
                            Default set to 0, so consider all candidates
        """
        self.logger.info("Mapping %i source terms...", len(source_terms))
        self.logger.info("...against %i ontology terms (%i labels/synonyms)", len(self.target_ontology_terms), len(self.target_labels))
        start = time.time()
        source_terms_norm = onto_utils.normalize_list(source_terms)
        vectorizer = self._tokenize(source_terms_norm, self.target_labels)
        results_mtx = self._sparse_dot_top(vectorizer, source_terms_norm, self.target_labels, min_score)
        results_df = self._get_mappings(results_mtx, max_mappings, source_terms, source_terms_ids, self.target_terms)
        end = time.time()
        self.logger.info("...done (mapping time: %.2fs seconds)", end-start)
        return results_df

    def _tokenize(self, source_terms, target_labels, analyzer='char_wb', n=3):
        """
        Tokenizes the (source) input strings and the target labels based on the selected analyzer
        :param source_terms: List of source terms to be matched
        :param target_labels: List of labels from ontology terms to be matched against
        :param analyzer: Type of analyzer ('char_wb', 'word')
        :param n: The gram length n (when using n-gram analyzer)
        :return TF-IDF Vectorizer
        """
        # Create count vectorizer and fit it on both lists to get vocabulary
        count_vectorizer = CountVectorizer(analyzer=analyzer, ngram_range=(n, n))
        vocabulary = count_vectorizer.fit(source_terms + target_labels).vocabulary_
        return TfidfVectorizer(vocabulary=vocabulary, analyzer=analyzer, ngram_range=(n, n))

    def _sparse_dot_top(self, vectorizer, source_terms, target_labels, min_score):
        src_mtx = vectorizer.fit_transform(source_terms).tocsr()
        tgt_mtx = vectorizer.fit_transform(target_labels).transpose().tocsr()
        # 'ntop' specifies the maximum number of labels/synonyms that should be considered
        # multiple labels/synonyms in the 'ntop' matches may be from the same ontology term
        return ct.awesome_cossim_topn(src_mtx, tgt_mtx, ntop=50, lower_bound=min_score)

    def _get_mappings(self, results_mtx, max_mappings, source_terms, source_terms_ids, target_terms):
        """ Build and return dataframe for mapping results along with term graphs for the obtained mappings """
        coo_mtx = results_mtx.tocoo()
        mappings = []
        last_source_term = ""
        top_mappings = set()
        for row, col, score in zip(coo_mtx.row, coo_mtx.col, coo_mtx.data):
            source_term = source_terms[row]
            source_term_id = source_terms_ids[row]
            onto_term = target_terms[col]
            self.logger.debug("Source term: %s maps to %s (%f)", source_term, onto_term.label, score)
            if source_term == last_source_term:
                if len(top_mappings) == max_mappings:
                    continue
            else:
                last_source_term = source_term
                top_mappings.clear()
            if onto_term.iri not in top_mappings:
                mappings.append(TermMapping(source_term, source_term_id, onto_term.label, onto_term.iri, score))
                top_mappings.add(onto_term.iri)
        return TermMappingCollection(mappings).mappings_df()

    def _get_target_labels_terms(self, ontology_terms):
        """Get lists of labels and terms to enable retrieving terms from their labels"""
        target_labels, target_terms = [], []
        for term in ontology_terms:
            for label in term.labels:
                target_labels.append(label)
                target_terms.append(term)
            for synonym in term.synonyms:
                target_labels.append(synonym)
                target_terms.append(term)
        return target_labels, target_terms



from text2term import onto_utils
from text2term.term_graph import TermGraph, Node, Edge
from owlready2 import Thing, ThingClass


class TermGraphGenerator:

    def __init__(self):
        pass

    def graph(self, term):
        """ Build and return a graph representing the neighborhood of an ontology term. """
        nodes, edges = set(), set()
        nodes.add(Node(term.iri, term.label))
        self._add_superclasses(term, nodes, edges)
        self._add_subclasses(term, term.children, nodes, edges)
        self._add_instances(term, term.instances, nodes, edges)
        return TermGraph(term.iri, nodes, edges)

    def _add_superclasses(self, term, nodes, edges):
        for parent in term.parents:
            self._add_node(parent, nodes)
            edges.add(Edge(term.iri, parent.iri, Edge.IS_A))
            self._add_ancestors(parent, nodes, edges)

    def _add_ancestors(self, node, nodes, edges):
        for ancestor in node.is_a:
            if ancestor is not Thing and isinstance(ancestor, ThingClass):
                self._add_node(ancestor, nodes)
                edges.add(Edge(node.iri, ancestor.iri, Edge.IS_A))
                self._add_ancestors(ancestor, nodes, edges)

    def _add_children(self, term, children, edge_type, nodes, edges):
        for child in children:
            self._add_node(child, nodes)
            edges.add(Edge(child.iri, term.iri, edge_type))

    def _add_subclasses(self, term, subclasses, nodes, edges):
        self._add_children(term, subclasses, Edge.IS_A, nodes, edges)

    def _add_instances(self, term, instances, nodes, edges):
        self._add_children(term, instances, Edge.INSTANCE_OF, nodes, edges)

    def _add_node(self, term, term_set):
        if len(term.label) == 0:
            label = onto_utils.label_from_iri(term.iri)
        else:
            label = term.label[0]
        term_set.add(Node(term.iri, label))

    def graphs_dicts(self, terms):
        """Convenience function to get a list of all term graphs' dictionary representations"""
        graph_dicts = []
        for term in terms:
            graph_dicts.append(self.graph(term).as_dict())
        return graph_dicts



"""Provides ZoomaMapper class"""

import json
import logging
import time
import requests
from text2term import onto_utils
from text2term.term_mapping import TermMappingCollection, TermMapping


class ZoomaMapper:

    def __init__(self):
        self.logger = onto_utils.get_logger(__name__, logging.INFO)
        self.url = "http://www.ebi.ac.uk/spot/zooma/v2/api/services/annotate"

    def map(self, source_terms, source_terms_ids, ontologies, max_mappings=3, api_params=()):
        """
        Find and return ontology mappings through the Zooma Web service
        :param source_terms: Collection of source terms to map to target ontologies
        :param source_terms_ids: List of identifiers for the given source terms
        :param ontologies: Comma-separated list of ontology acronyms (eg 'HP,EFO') or 'all' to search all ontologies
        :param max_mappings: The maximum number of (top scoring) ontology term mappings that should be returned
        :param api_params: Additional Zooma API-specific parameters to include in the request
        """
        self.logger.info("Mapping %i source terms against ontologies: %s...", len(source_terms), ontologies)
        start = time.time()
        mappings = []
        for term, term_id in zip(source_terms, source_terms_ids):
            mappings.extend(self._map_term(term, term_id, ontologies, max_mappings, api_params))
        self.logger.info('done (mapping time: %.2fs seconds)', time.time()-start)
        return TermMappingCollection(mappings).mappings_df()

    def _map_term(self, source_term, source_term_id, ontologies, max_mappings, api_params):
        # see https://www.ebi.ac.uk/spot/zooma/docs/api for details of API parameters
        params = {
            "propertyValue": onto_utils.normalize(source_term),
            "filter": "required:[gwas,cttv,atlas,eva-clinvar,sysmicro],ontologies:[" + ontologies + "]"
        }
        if len(api_params) > 0:
            params.update(api_params)
        self.logger.debug("API parameters: " + str(params))
        mappings = []
        self.logger.debug("Searching for ontology terms to match: " + source_term)
        response = self._do_get_request(self.url, params=params)
        if response is not None:
            self.logger.debug("...found " + str(len(response)) + " mappings")
            for mapping in response:
                if len(mappings) < max_mappings:
                    mappings.append(self._mapping_details(source_term, source_term_id, mapping))
        return mappings

    def _mapping_details(self, source_term, source_term_id, mapping_response):
        # get ontology term label
        ann_class = mapping_response["annotatedProperty"]
        term_label = ann_class["propertyValue"]

        # get ontology term IRI
        tags = mapping_response["semanticTags"]
        term_iri = tags[0]

        mapping_score = self._mapping_score(mapping_response["confidence"])
        return TermMapping(source_term, source_term_id, term_label, term_iri, mapping_score)

    def _mapping_score(self, confidence):
        """Represent numerically the mapping confidence categories returned by Zooma (high, good, medium or low)"""
        if confidence == "HIGH":
            return 1.0
        elif confidence == "GOOD":
            return 0.75
        elif confidence == "MEDIUM":
            return 0.5
        elif confidence == "LOW":
            return 0.25
        else:
            return 0

    def _do_get_request(self, request_url, params=None):
        response = requests.get(request_url, params=params, verify=True)
        if response.ok:
            json_resp = json.loads(response.content)
            if len(json_resp) > 0:
                return json_resp
            else:
                self.logger.info("Empty response for input: " + request_url + " with parameters " + str(params))
        else:
            json_resp = json.loads(response.content)
            self.logger.error(response.reason + ":" + request_url + ". " + json_resp["errors"][0])



from setuptools import setup, find_packages

version = '0.4.1'
description = 'A tool for mapping free-text descriptions of (biomedical) entities to controlled terms in an ontology'
long_description = open('README.md').read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='text2term',
    version=version,
    install_requires=requirements,
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/ccb-hms/ontology-mapper',
    license='MIT',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Center for Computational Biomedicine, Harvard Medical School',
    author_email='rafael_goncalves@hms.harvard.edu',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    python_requires=">=3.9",
)



"""Provides OntologyTerm class"""


class OntologyTerm:

    def __init__(self, iri, labels, synonyms, definition, parents=(), children=(), instances=()):
        self._iri = iri
        self._labels = labels
        self._synonyms = synonyms
        self._definition = definition
        self._parents = parents
        self._children = children
        self._instances = instances

    @property
    def iri(self):
        return self._iri

    @property
    def labels(self):
        return self._labels

    @property
    def synonyms(self):
        return self._synonyms

    @property
    def definition(self):
        return self._definition

    @property
    def parents(self):
        return self._parents

    @property
    def children(self):
        return self._children

    @property
    def instances(self):
        return self._instances

    @property
    def label(self):
        """Return a single label for this term"""
        return next(iter(self.labels))

    def __eq__(self, other):
        if isinstance(other, OntologyTerm):
            return self._iri == other._iri
        return False

    def __hash__(self):
        return hash(str(self._iri))

    def __str__(self):
        return "Ontology Term: " + self.iri + ", Labels: " + str(self.labels) + ", Synonyms: " + \
               str(self.synonyms) + ", Definition: " + str(self.definition) + ", Parents: " + str(self.parents) + \
               ", Children: " + str(self.children) + ", Instances: " + str(self.instances)
